<?php

declare(strict_types=1);

/*
 * Copyright by SoftCreatR Media.
 *
 * License: https://www.softcreatr.com/license-terms
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * The above copyright notice and this disclaimer notice shall be included in all
 * copies or substantial portions of the Software.
 */

namespace wcf\system\event\listener;

use Closure;
use Generator;
use GuzzleHttp\{Pool, Psr7\Response, RequestOptions};
use Throwable;
use wcf\data\package\PackageCache;
use wcf\system\io\http\ScRedirectGuard;
use wcf\system\{application\ApplicationHandler, exception\SystemException, io\HttpFactory, worker\AbstractWorker};
use wcf\util\{ArrayUtil, MessageUtil, StringUtil, Url};

use function basename;
use function chr;
use function class_exists;
use function GuzzleHttp\Psr7\stream_for;
use function hexdec;
use function mb_detect_encoding;
use function mb_strrpos;
use function mb_strtolower;
use function preg_match;
use function preg_replace_callback;
use function urldecode;

class ReplaceURLByNameListener implements IParameterizedEventListener
{
    /**
     * @var integer
     */
    const MAX_LENGTH = 1 << 20; // 1 MiB

    /**
     * @var array
     */
    private $titles;

    /**
     * @var array
     */
    private $blockList;

    /**
     * @inheritdoc
     */
    public function execute($eventObj, $className, $eventName, array &$parameters)
    {
        // do nothing, if WSC >= 5.4 and URL unfurling is enabled
        if (defined('MODULE_URL_UNFURLING') && MODULE_URL_UNFURLING) {
            return;
        }

        // do nothing, if this listener is invoked from a worker
        if (class_exists(AbstractWorker::class, false)) {
            return;
        }

        $elements = [];

        foreach ($eventObj->getDocument()->getElementsByTagName('a') as $element) {
            $url = $element->getAttribute('href');

            if ($url === $element->textContent) {
                $urlComponents = Url::parse($url);

                if (
                    empty($urlComponents['host']) ||
                    empty($urlComponents['scheme']) ||
                    !empty($urlComponents['port']) ||
                    0 !== strncmp($urlComponents['scheme'], 'http', 4) ||
                    $this->isBlocked($urlComponents['host'])
                ) {
                    continue;
                }

                $elements[] = $element;
            }
        }

        if (!empty($elements)) {
            try {
                $this->processURLs($elements);
            } catch (Throwable $e) {
                return;
            }
        }
    }

    /**
     * Performs connections and DOM manipulations.
     */
    private function processURLs(array $elements = [])
    {
        $numUrls = count($elements);
        $ua = HttpFactory::getDefaultUserAgent();

        if ('extended' === REPLACE_URL_BY_NAME_USER_AGENT) {
            try {
                $packageVersion = PackageCache::getInstance()
                    ->getPackageByIdentifier('de.softcreatr.wsc.replaceurlbyname')
                    ->packageVersion;

                $ua = 'Mozilla/5.0 (' . $ua . ' ReplaceUrlByName/' . $packageVersion . ')';
            } catch (SystemException $e) {
                // ignore
            }
        }

        $client = HttpFactory::makeClient(
            [
                RequestOptions::TIMEOUT => REPLACE_URL_BY_NAME_TIMEOUT,
                RequestOptions::HEADERS => [
                    'User-Agent' => $ua,
                    'Range' => 'bytes=0-' . (self::MAX_LENGTH - 1)
                ],
                RequestOptions::STREAM => true,
                RequestOptions::SINK => stream_for(fopen("php://temp", "wb+")),
                RequestOptions::ALLOW_REDIRECTS => REPLACE_URL_BY_NAME_FOLLOW_REDIRECTS ? [
                    'max' => REPLACE_URL_BY_NAME_FOLLOW_MAX_REDIRECTS,
                    'on_redirect' => new ScRedirectGuard(),
                    'protocols' => ['http', 'https'],
                    'track_redirects' => true
                ] : false
            ]
        );

        // single URL, no async needed
        if (1 === $numUrls) {
            try {
                $url = $elements[0]->getAttribute('href');
                $response = $client->get($url);

                $title = $this->fetchTitle($url, $response);

                if (!empty($title)) {
                    $elements[0]->textContent = $title;
                }
            } catch (Throwable $e) {
                // ignore
            }

            return;
        }

        // async requests
        $requests = static function($total) use ($client, $elements): Generator {
            $index = 0;

            for ($iter = 0; $iter < $total; $iter++) {
                $url = StringUtil::trim($elements[$iter]->getAttribute('href'));

                yield $index => static function() use ($client, $url, &$index) {
                    $index++;

                    return $client->getAsync($url);
                };
            }
        };

        foreach (Pool::batch($client, $requests($numUrls < 100 ? $numUrls : 100)) as $pool => $response) {
            if ($response instanceof Throwable) {
                continue;
            }

            $title = $this->fetchTitle($elements[$pool]->getAttribute('href'), $response);

            if (!empty($title)) {
                $elements[$pool]->textContent = $title;
            }
        }
    }

    /**
     * Returns the content of the Title-Tag from a given URL.
     */
    private function fetchTitle(string $url, Response $response): string
    {
        $title = $charset = $mimeType = '';
        $urlLower = mb_strtolower($url);
        $body = '';
        $bodyLength = 0;

        // return title from "cache", if the given URL has been processed in the past
        if (isset($this->titles[$urlLower])) {
            return $this->titles[$urlLower];
        }

        while (!$response->getBody()->eof()) {
            $toRead = min(8192, self::MAX_LENGTH - $bodyLength);
            $data = $response->getBody()->read($toRead);
            $body .= $data;
            $bodyLength += strlen($data);

            if ($bodyLength >= self::MAX_LENGTH) {
                break;
            }
        }

        $response->getBody()->close();

        // determine mime-type and charset
        if ($response->hasHeader('content-type')) {
            foreach ($response->getHeader('content-type') as $header) {
                if (preg_match('~([\w/+]+)(?:;\s+charset=(\S+))?~i', $header, $matches)) {
                    if (!empty($matches[1])) {
                        $mimeType = $matches[1];
                    }

                    if (!empty($matches[2]) && in_array($matches[2], mb_list_encodings(), true)) {
                        $charset = strtoupper($matches[2]);
                    }
                }
            }
        }

        // title tags should just appear in text/html documents
        if (null !== $body && ('text/html' === $mimeType || 'application/xhtml+xml' === $mimeType)) {
            if (
                preg_match('/<title[^>]*>(.*?)<\/title[^>]*>/is', $body, $remoteTitle) &&
                !empty($remoteTitle[1])
            ) {
                $title = StringUtil::trim($remoteTitle[1]);
            }

            // if no charset has been found, try to get it somewhere else
            if (empty($charset)) {
                $charset = strtoupper(mb_detect_encoding($title, mb_list_encodings(), true));
            }

            // convert encoding, if needed
            if (!empty($charset) && 'UTF-8' !== $charset && in_array($charset, mb_list_encodings(), true)) {
                try {
                    $title = StringUtil::convertEncoding($charset, 'UTF-8', $title);
                } catch (Throwable $e) {
                    // ignore
                }
            }
        }

        // no ability to get page title, fallback to filename
        if (empty($title)) {
            $title = basename(urldecode($url));

            // unable to get page title and file doesn't seem to have
            // a proper filename, so return the original url instead
            if (false === mb_strrpos($title, '.')) {
                $this->titles[$url] = $url;

                return $url;
            }
        }

        // return new link title
        $title = StringUtil::decodeHTML(StringUtil::trim(MessageUtil::stripCrap($title)));

        // due to the fact, that StringUtil can't handle hex
        // encoded html entities, we'll have to do this by ourselves
        $title = preg_replace_callback('/&#x([a-f0-9]+)(;)/mi', static function($matches) {
            return chr(hexdec($matches[1]));
        }, $title);

        $this->titles[$url] = $title;

        return $title;
    }

    /**
     * Returns a function that matches hosts against the given blocklist.
     * The blocklist supports wildcards using using `*.` prefix.
     */
    private function getHostMatcher(array $blocklist): Closure
    {
        $hosts = [];

        foreach ($blocklist as $host) {
            $isWildcard = false;

            if (false !== mb_strpos($host, '*')) {
                $host = preg_replace('~^(\*\.)+~', '', $host);

                if ('' === $host || mb_strpos($host, '*') !== false) {
                    // bad host
                    continue;
                }

                $isWildcard = true;
            }

            $host = mb_strtolower($host);

            if (!isset($hosts[$host])) {
                $hosts[$host] = $isWildcard;
            }
        }

        return static function ($hostname) use ($hosts) {
            static $validHosts = [];

            $hostname = mb_strtolower($hostname);

            if (isset($hosts[$hostname]) || isset($validHosts[$hostname])) {
                return true;
            }

            // check wildcard hosts
            foreach ($hosts as $host => $isWildcard) {
                // the prepended dot will ensure that `example.com` matches only
                // on domains like `foo.example.com` but not on `bar-example.com`
                if (
                    $isWildcard &&
                    false !== mb_strpos($hostname, $host) &&
                    StringUtil::endsWith($hostname, '.' . $host)
                ) {
                    $validHosts[$hostname] = $hostname;

                    return true;
                }
            }

            return false;
        };
    }

    /**
     * Returns whether the given URLs hostname should be processed, or not.
     */
    private function isBlocked(string $hostname): bool
    {
        static $matcher = null;

        if (null === $matcher) {
            if (null === $this->blockList) {
                $this->blockList = ArrayUtil::trim(
                    explode("\n", mb_strtolower(StringUtil::unifyNewlines(REPLACE_URL_BY_NAME_BLOCKLIST)))
                );

                if (REPLACE_URL_BY_NAME_BLOCK_INTERNAL_URLS) {
                    try {
                        foreach (ApplicationHandler::getInstance()->getApplications() as $application) {
                            $host = mb_strtolower($application->domainName);
                            $this->blockList[] = $host;
                        }
                    } catch (SystemException $e) {
                        // ignore
                    }
                }
            }

            $matcher = $this->getHostMatcher($this->blockList);
        }

        return $matcher($hostname);
    }
}
